/* 
   Unix SMB/Netbios implementation.
   Version 1.9.
   SMB parameters and setup
   Copyright (C) Andrew Tridgell 1992-1997
   Copyright (C) Luke Kenneth Casson Leighton 1996-1997
   Copyright (C) Paul Ashton 1997
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#ifndef _RPC_WKS_H /* _RPC_WKS_H */
#define _RPC_WKS_H 


/* wkssvc pipe */
#define WKS_QUERY_INFO      0x00
#define WKS_USER_ENUM       0x02
#define WKS_TRANSPORT_ENUM  0x05
#define WKS_UNJOIN          0x17


/* WKS_Q_QUERY_INFO - probably a capabilities request */
typedef struct q_wks_query_info_info
{
	uint32 ptr_srv_name;         /* pointer (to server name?) */
	UNISTR2 uni_srv_name;        /* unicode server name starting with '\\' */

	uint16 switch_value;            /* info level 100 (0x64) */

} WKS_Q_QUERY_INFO;


/* WKS_INFO_100 - level 100 info */
typedef struct wks_info_100_info
{
	uint32 platform_id;          /* 0x0000 01f4 - unknown */
	uint32 ptr_compname;       /* pointer to server name */
	uint32 ptr_lan_grp ;       /* pointer to domain name */
	uint32 ver_major;          /* 4 - unknown */
	uint32 ver_minor;          /* 0 - unknown */

	UNISTR2 uni_compname;      /* unicode server name */
	UNISTR2 uni_lan_grp ;      /* unicode domain name */

} WKS_INFO_100;


/* WKS_R_QUERY_INFO - probably a capabilities request */
typedef struct r_wks_query_info_info
{
	uint16 switch_value;          /* 100 (0x64) - switch value */

	/* for now, only level 100 is supported.  this should be an enum container */
	uint32 ptr_1;              /* pointer 1 */
	WKS_INFO_100 *wks100;      /* workstation info level 100 */

	uint32 status;             /* return status */

} WKS_R_QUERY_INFO;


/* WKS_Q_USER_ENUM */
typedef struct
{
	uint32 ptr_srv_name;         /* pointer (to server name?) */
	UNISTR2 uni_srv_name;        /* unicode server name starting with '\\' */

	uint32 unk0; /* 0x0 */
	uint32 unk1; /* 0x0 */
	uint32 unk_ptr; /* 0x1 */
	uint32 unk2; /* 0x0 */
	uint32 unk3; /* 0x0 */
	uint32 max_size;  /* max returned size 0xffffffff okay */
	uint32 unk4; /* 0x0 */

} WKS_Q_USER_ENUM;

typedef struct
{
	uint32 ptr_username;
	UNISTR2 uni_username;
	uint32 ptr_domain;
	UNISTR2 uni_domain;
	uint32 ptr_unk;
	UNISTR2 uni_unk;
	uint32 ptr_logonserver;
	UNISTR2 uni_logonserver;
} WKS_LOGGED_IN_USER;

typedef struct
{
	uint32 level1;
	uint32 level2;
	uint32 ptr1;
	uint32 count1;
	uint32 ptr2;
	uint32 count2;
	WKS_LOGGED_IN_USER *users;
	uint32 count3;
	uint32 unk;
	uint32 status;
} WKS_R_USER_ENUM;


typedef struct
{
	uint32 ptr_srv_name;
	UNISTR2 uni_srv_name;

	uint32 unk0; /* 0 */
	uint32 unk1; /* 0 */
	uint32 unk2; /* 4 */
} WKS_Q_UNJOIN;

typedef struct
{
	uint32 status;
} WKS_R_UNJOIN;




#endif /* _RPC_WKS_H */
