/* 
   Unix SMB/Netbios implementation.
   Version 1.9.
   SMB parameters and setup
   Copyright (C) Elrond                            2002
   Copyright (C) Andrew Tridgell              1992-2000
   Copyright (C) John H Terpstra              1996-2000
   Copyright (C) Luke Kenneth Casson Leighton 1996-2001
   Copyright (C) Paul Ashton                  1998-2000
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#ifndef _SAMDB_H
#define _SAMDB_H


#ifdef HAVE_GRP_H
#include <grp.h>
#endif

#if defined(WITH_LDAP)
#include <lber.h>
#include <ldap.h>
#else
#define LDAPMod void
#endif

#include "rpc_lsa.h"
#include "rpc_samr.h"
#include "ldapdb.h"


struct sam_passwd
{
	NTTIME logon_time;	/* logon time */
	NTTIME logoff_time;	/* logoff time */
	NTTIME kickoff_time;	/* kickoff time */
	NTTIME pass_last_set_time;	/* password last set time */
	NTTIME pass_can_change_time;	/* password can change time */
	NTTIME pass_must_change_time;	/* password must change time */

	char *unix_name;	/* unix username string */
	char *nt_name;		/* nt username string */
	char *full_name;	/* user's full name string */
	char *home_dir;		/* home directory string */
	char *dir_drive;	/* home directory drive string */
	char *logon_script;	/* logon script string */
	char *profile_path;	/* profile path string */
	char *acct_desc;	/* user description string */
	char *workstations;	/* login from workstations string */
	char *unknown_str;	/* don't know what this is, yet. */
	char *munged_dial;	/* munged path name and dial-back tel number */

	uid_t unix_uid;		/* this is actually the unix uid_t */
	uint32 user_rid;	/* Primary User ID */
	uint32 group_rid;	/* Primary Group ID */

	unsigned char *smb_passwd;	/* Null if no password */
	unsigned char *smb_nt_passwd;	/* Null if no password */

	uint32 acct_ctrl;	/* account info (ACB_xxxx bit-mask) */
	uint32 unknown_3;	/* 0x00ff ffff */

	uint16 logon_divs;	/* 168 - number of hours in a week */
	uint32 hours_len;	/* normally 21 bytes */
	uint8 hours[MAX_HOURS_LEN];
};

struct smb_passwd
{
	uid_t unix_uid;		/* unix userid */
	char *unix_name;	/* unix username string */

	uint32 user_rid;	/* Primary User ID */
	char *nt_name;		/* unix username string */

	unsigned char *smb_passwd;	/* Null if no password */
	unsigned char *smb_nt_passwd;	/* Null if no password */

	uint32 acct_ctrl;	/* account info (ACB_xxxx bit-mask) */
	time_t pass_last_set_time;	/* password last set time */
};


/*
 * Each implementation of the password database code needs
 * to support the following operations.
 */

struct smb_passdb_ops
{
	/*
	 * Password database operations.
	 */
	char *dead11;

	char *dead1;
	char *dead2;
	char *dead3;
	char *dead4;

	/*
	 * smb password database query functions.
	 */
	char *dead7;
	char *dead5;
	char *dead6;

	/*
	 * smb password database modification functions.
	 */
	char *dead8;
	char *dead9;
	char *dead10;
};

/*
 * Each implementation of the password database code needs
 * to support the following operations.
 */

struct sam_passdb_ops
{
	/*
	 * Password database operations.
	 */
	void *(*startsam21pwent) (BOOL);
	void (*endsam21pwent) (void *);
	char *dead1;
	char *dead2;

	/*
	 * sam password database query functions.
	 */
	struct sam_passwd *(*getsam21pwntnam) (const char *);
	char *dead3;
	struct sam_passwd *(*getsam21pwrid) (uint32);
	struct sam_passwd *(*getsam21pwent) (void *);

	/*
	 * sam password database modification functions.
	 */
	BOOL (*add_sam21pwd_entry) (struct sam_passwd *);
	BOOL (*mod_sam21pwd_entry) (struct sam_passwd *, BOOL);
	BOOL (*del_sam21pwd_entry) (uint32 rid);
	uint32 (*getsamseqnum) (void);

	void * (* open_simple) (void);
	void (* close_simple) (void *);
};

/*
 * Each implementation of the group database code needs
 * to support the following operations.
 *
 * This allows enumeration, modification and addition of groups.  there
 * is _no_ deletion of groups: you can only modify them to a status of
 * "deleted" (this by the way is a requirement of c2 rating)
 */

struct groupdb_ops
{
	/*
	 * Group database ops.
	 */
	void *(*startgroupent) (BOOL);
	void (*endgroupent) (void *);
	char *dead1;
	char *dead2;

	/*
	 * group database query functions. 
	 */
	DOMAIN_GRP *(*getgroupntnam) (const char *, DOMAIN_GRP_MEMBER **,
				      int *);
	char *dead3;
	DOMAIN_GRP *(*getgrouprid) (uint32, DOMAIN_GRP_MEMBER **, int *);
	DOMAIN_GRP *(*getgroupent) (void *, DOMAIN_GRP_MEMBER **, int *);

	/*
	 * group database modification functions.
	 */
	BOOL (*add_group_entry) (DOMAIN_GRP *);
	BOOL (*mod_group_entry) (DOMAIN_GRP *);
	BOOL (*del_group_entry) (uint32);

	BOOL (*add_group_member) (uint32, uint32);
	BOOL (*del_group_member) (uint32, uint32);

	/*
	 * user group functions
	 */
	BOOL (*getusergroupsntnam) (const char *, DOMAIN_GRP **, int *);
};

/*
 * Each implementation of the alias database code needs
 * to support the following operations.
 *
 * This allows enumeration, modification and addition of aliases.  there
 * is _no_ deletion of aliases: you can only modify them to a status of
 * "deleted" (this by the way is a requirement of c2 rating)
 */

struct aliasdb_ops
{
	/*
	 * Alias database ops.
	 */
	void *(*startaliasent) (BOOL);
	void (*endaliasent) (void *);
	char *dead1;
	char *dead2;

	/*
	 * alias database query functions. 
	 */
	LOCAL_GRP *(*getaliasntnam) (const char *, LOCAL_GRP_MEMBER **,
				     int *);
	char *dead3;
	LOCAL_GRP *(*getaliasrid) (uint32, LOCAL_GRP_MEMBER **, int *);
	LOCAL_GRP *(*getaliasent) (void *, LOCAL_GRP_MEMBER **, int *);

	/*
	 * alias database modification functions.
	 */
	BOOL (*add_alias_entry) (LOCAL_GRP *);
	BOOL (*mod_alias_entry) (LOCAL_GRP *);
	BOOL (*del_alias_entry) (uint32);

	BOOL (*add_alias_member) (uint32, const DOM_SID *);
	BOOL (*del_alias_member) (uint32, const DOM_SID *);

	/*
	 * user alias functions
	 */
	BOOL (*getuseraliasntnam) (const char *, LOCAL_GRP **, int *);
};


/*The following definitions come from  groupdb/aliasdb.c  */

BOOL initialise_alias_db(void);
LOCAL_GRP *iterate_getaliasrid(uint32 rid, LOCAL_GRP_MEMBER **mem, int *num_mem);
LOCAL_GRP *iterate_getaliasntnam(const char *name, LOCAL_GRP_MEMBER **mem, int *num_mem);
BOOL add_alias_to_list(LOCAL_GRP **alss, int *num_alss, const LOCAL_GRP *als);
BOOL als_user_is_member(const char *user_name, LOCAL_GRP_MEMBER *mem, int num_mem);
BOOL iterate_getuseraliasntnam(const char *user_name, LOCAL_GRP **alss, int *num_alss);
BOOL enumdomaliases(LOCAL_GRP **alss, int *num_alss);
void *startaliasent(BOOL update);
void endaliasent(void *vp);
LOCAL_GRP *getaliasent(void *vp, LOCAL_GRP_MEMBER **mem, int *num_mem);
BOOL add_alias_entry(LOCAL_GRP *newgrp);
BOOL mod_alias_entry(LOCAL_GRP* als);
BOOL del_alias_entry(uint32 rid);
BOOL add_alias_member(uint32 rid, const DOM_SID *member_sid);
BOOL del_alias_member(uint32 rid, const DOM_SID *member_sid);
LOCAL_GRP *getaliasntnam(const char *name, LOCAL_GRP_MEMBER **mem, int *num_mem);
LOCAL_GRP *getaliasrid(uint32 alias_rid, LOCAL_GRP_MEMBER **mem, int *num_mem);
BOOL getuseraliasntnam(const char *user_name, LOCAL_GRP **als, int *num_alss);
void aldb_init_als(LOCAL_GRP *als);
BOOL make_alias_line(char *p, int max_len,
				LOCAL_GRP *als,
				LOCAL_GRP_MEMBER **mem, int *num_mem);

/*The following definitions come from  groupdb/builtindb.c  */

BOOL initialise_builtin_db(void);
LOCAL_GRP *iterate_getbuiltinrid(uint32 rid, LOCAL_GRP_MEMBER **mem, int *num_mem);
LOCAL_GRP *iterate_getbuiltinntnam(const char *name, LOCAL_GRP_MEMBER **mem, int *num_mem);
BOOL iterate_getuserbuiltinntnam(const char *user_name, LOCAL_GRP **blts, int *num_blts);
BOOL enumdombuiltins(LOCAL_GRP **blts, int *num_blts);
void *startbuiltinent(BOOL update);
void endbuiltinent(void *vp);
LOCAL_GRP *getbuiltinent(void *vp, LOCAL_GRP_MEMBER **mem, int *num_mem);
BOOL add_builtin_entry(LOCAL_GRP *newblt);
BOOL mod_builtin_entry(LOCAL_GRP* blt);
BOOL add_builtin_member(uint32 rid, const DOM_SID *member_sid);
BOOL del_builtin_member(uint32 rid, const DOM_SID *member_sid);
LOCAL_GRP *samdb_builtin_find(const char *name);
LOCAL_GRP *getbuiltinrid(uint32 builtin_rid, LOCAL_GRP_MEMBER **mem, int *num_mem);
BOOL getuserbuiltinntnam(const char *user_name, LOCAL_GRP **blt, int *num_blts);
void bidb_init_blt(LOCAL_GRP *blt);

/*The following definitions come from  groupdb/groupdb.c  */

BOOL initialise_group_db(void);
DOMAIN_GRP *iterate_getgrouprid(uint32 rid, DOMAIN_GRP_MEMBER **mem, int *num_mem);
DOMAIN_GRP *iterate_getgroupntnam(const char *name, DOMAIN_GRP_MEMBER **mem, int *num_mem);
BOOL iterate_getusergroupsnam(const char *user_name, DOMAIN_GRP **grps, int *num_grps);
BOOL enumdomgroups(DOMAIN_GRP **grps, int *num_grps);
void *startgroupent(BOOL update);
void endgroupent(void *vp);
DOMAIN_GRP *getgroupent(void *vp, DOMAIN_GRP_MEMBER **mem, int *num_mem);
BOOL add_group_entry(DOMAIN_GRP *newgrp);
BOOL del_group_entry(uint32 rid);
BOOL mod_group_entry(DOMAIN_GRP* grp);
BOOL add_group_member(uint32 rid, uint32 member_rid);
BOOL del_group_member(uint32 rid, uint32 member_rid);
DOMAIN_GRP *getgroupntnam(const char *name);
DOMAIN_GRP *getgrouprid(uint32 group_rid, DOMAIN_GRP_MEMBER **mem, int *num_mem);
BOOL samdb_getusergroups(struct sam_passwd *sam_pass,
			 DOMAIN_GRP **grps, size_t *num_grps);
void gpdb_init_grp(DOMAIN_GRP *grp);
BOOL make_group_line(char *p, int max_len,
				DOMAIN_GRP *grp,
				DOMAIN_GRP_MEMBER **mem, int *num_mem);

/*The following definitions come from  *file.c  */

struct groupdb_ops *file_initialise_group_db(void);
struct aliasdb_ops *file_initialise_alias_db(void);

/*The following definitions come from  *ldap.c  */

struct groupdb_ops *ldap_initialise_group_db(void);
struct aliasdb_ops *ldap_initialise_alias_db(void);
struct aliasdb_ops *ldap_initialise_builtin_db(void);

/*The following definitions come from  *nt5ldap.c  */

BOOL nt5ldap_make_local_grp(LDAPDB * hds, LOCAL_GRP * group,
		     LOCAL_GRP_MEMBER ** members, int *num_membs, uint32 req_type);
BOOL nt5ldap_make_sam_user_info21(LDAPDB *hds, SAM_USER_INFO_21 *usr);
BOOL nt5ldap_sam_user_info21_mods(const SAM_USER_INFO_21 *usr, LDAPMod ***mods, int op,
	char *rdn, size_t rdnmaxlen, BOOL *iscomputer_p);
struct groupdb_ops *nt5ldap_initialise_group_db (void);
struct aliasdb_ops *nt5ldap_initialise_alias_db (void);
struct aliasdb_ops *nt5ldap_initialise_builtin_db (void);
struct sam_passdb_ops *nt5ldap_initialise_sam_password_db (void);
struct smb_passdb_ops *nt5ldap_initialise_password_db (void);

/*The following definitions come from  *unix.c  */

BOOL get_unixgroup_members(struct group *grp,
				int *num_mem, DOMAIN_GRP_MEMBER **members);
struct groupdb_ops *unix_initialise_group_db(void);
BOOL get_unixalias_members(struct group *grp,
				int *num_mem, LOCAL_GRP_MEMBER **members);
struct aliasdb_ops *unix_initialise_alias_db(void);
struct aliasdb_ops *unix_initialise_builtin_db(void);


/*The following definitions come from  passdb/sampassdb.c  */

BOOL initialise_sam_password_db(void);
void *samdb_open_simple(void);
void samdb_close_simple(void *open_ptr);
void *startsam21pwent(void);
void endsam21pwent(void *vp);
struct sam_passwd *getsam21pwent(void *vp);
BOOL add_sam21pwd_entry(struct sam_passwd *newpwd);
BOOL mod_sam21pwd_entry(struct sam_passwd* pwd, BOOL override);
struct sam_passwd *getsam21pwntnam(const char *name);
struct sam_passwd *getsam21pwrid(uint32 rid);
void pwdb_init_sam(struct sam_passwd *user);
void copy_id23_to_sam_passwd(struct sam_passwd *to, const SAM_USER_INFO_23 *from);
void copy_id21_to_sam_passwd(struct sam_passwd *to, const SAM_USER_INFO_21 *from);
void copy_sam_passwd(struct sam_passwd *to, const struct sam_passwd *from);
struct smb_passwd *pwdb_sam_to_smb(struct sam_passwd *user);
struct sam_passwd *pwdb_smb_to_sam(struct smb_passwd *user);
void sampwd_reset_pass_last_set_time(struct sam_passwd *sam);

/*The following definitions come from  passdb/passdb.c  */

BOOL initialise_password_db(void);
uint32 getsamseqnum(void);
BOOL del_smbpwd_entry(uint32 rid);
void pwdb_init_smb(struct smb_passwd *user);
struct smb_passwd *pwdb_smb_map_names(struct smb_passwd *smb);

/*The following definitions come from  passdb/smbpass.c  */

struct smb_passdb_ops *file_initialise_password_db(void);
struct sam_passdb_ops *file_initialise_sam_password_db(void);

/*The following definitions come from  *ldap.c  */

BOOL tng_ldap_init_basic(void);
BOOL ldap_connect(void);
void ldap_disconnect(void);
BOOL ldap_search_for(char *filter);
BOOL tng_ldap_open_dn(const char *dn, char **attrs);
char *tng_ldap_filter_quote(const char *filter_part);
BOOL ldap_get_attribute(const char *attribute, char *value);
void ldap_make_mod(LDAPMod ***modlist,int modop, char *attribute, char *value);
BOOL tng_ldap_commit_obj(LDAPMod **mods);
BOOL ldap_commit_mods(const char *attribute, const char *value,
		      LDAPMod **mods, BOOL add, const char *subcontainer);
BOOL ldap_search_by_rid(uint32 rid);
BOOL samldap_search_user_by_ntname(const char *ntname);
BOOL samldap_search_obj_by_rid(uint32 rid, const char *objclass);
BOOL ldap_allocaterid(uint32 *rid);
BOOL ldap_incr_seq_num(void);
struct sam_passdb_ops *ldap_initialise_sam_password_db(void);
struct smb_passdb_ops *ldap_initialise_password_db(void);

/*The following definitions come from  passdb/mysql*.c  */

int mysql_db_lock_connect( MYSQL *handle );
void *mysql_startpwent( BOOL update );
void mysql_endpwent( void *ptr );
void *mysql_fill_smb_passwd( MYSQL_ROW *row );
void *mysql_fetch_passwd( void *(*filler)(MYSQL_ROW*), char *where );
void *mysql_getpwnam(void *(*filler)(MYSQL_ROW *), char *field, const char *name);
struct smb_passwd *mysql_getsmbpwnam(const char *unix_name);
BOOL mysql_del_smb( MYSQL *handle, char *unix_name );
BOOL mysql_add_smb( MYSQL *handle, struct smb_passwd *smb );
BOOL mysql_mod_smb( MYSQL *handle, struct smb_passwd *smb, BOOL override );
BOOL mysql_add_smbpwd_entry(struct smb_passwd *smb);
BOOL mysql_mod_smbpwd_entry(struct smb_passwd *smb, BOOL override);
struct smb_passdb_ops *mysql_initialise_password_db(void);
struct sam_passdb_ops *mysql_initialise_sam_password_db(void);

/*The following definitions come from  passdb/nispass.c  */

struct smb_passdb_ops *nisplus_initialise_password_db(void);


#endif /* _SAMDB_H */
