#ifndef TNG_SMBD_H
#define TNG_SMBD_H

#ifdef HAVE_PWD_H
#include <pwd.h>
#else
struct passwd;  /* forward decl to calm compiler */
#endif

#include "tdb.h"
#include "ntdomain.h"
#include "rpc_secdes.h"
#include "msdfs.h"

struct pipes_struct;

/**
 * A connection by a client
 * Mostly a "tcp connection"
 */
typedef struct _smbd_conn {
	int socket;
	char *client_addr;
	char *client_name;
	uint32 client_caps;
	BOOL in_oplock_break;
	BOOL client_failed_oplock_break;
	/* 
	 * variables used to override error codes. 
	 */
	int unix_ERR_class;
	int unix_ERR_code;
	NTSTATUS unix_ERR_ntstatus;
} SmbdConn;


typedef struct {
	NTTIME CreationTime;
	NTTIME LastAccessTime;
	NTTIME LastWriteTime;
	NTTIME ChangeTime;
	SMB_OFF_T AllocationSize;
	SMB_OFF_T EndOfFile;
	uint32 FileAttributes;
} NTFileStatInfo;


/* Structure used when SMBwritebmpx is active */
typedef struct
{
	size_t wr_total_written;	/* So we know when to discard this */
	int32 wr_timeout;
	int32 wr_errclass;
	int32 wr_error;		/* Cached errors */
	BOOL wr_mode;		/* write through mode) */
	BOOL wr_discard;	/* discard all further data */
} write_bmpx_struct;

typedef struct write_cache
{
	SMB_OFF_T file_size;
	SMB_OFF_T offset;
	size_t alloc_size;
	size_t data_size;
	char *data;
} write_cache;

typedef struct files_struct
{
	struct files_struct *next, *prev;
	int fnum;
	struct connection_struct *conn;
	int fd;
	int print_jobid;
	SMB_DEV_T dev;
	SMB_INO_T inode;
	BOOL delete_on_close;
	SMB_OFF_T pos;
	SMB_OFF_T size;
	mode_t mode;
	uint16 vuid;
	uint16 open_smbpid; /* the smbpid, that opened this file */
	uint32 desired_access;
	write_bmpx_struct *wbmpx_ptr;
	write_cache *wcp;
	struct timeval open_time;
	int share_mode;
	time_t pending_modtime;
	int oplock_type;
	int sent_oplock_break;
	BOOL can_lock;
	BOOL can_read;
	BOOL can_write;
	BOOL can_execute;
	BOOL print_file;
	BOOL modified;
	BOOL is_directory;
	BOOL directory_delete_on_close;
	BOOL stat_open;
	char *dos_fsp_name;
	struct pipes_struct *pipe_info;
} files_struct;

typedef struct pipes_struct
{
	struct pipes_struct *next, *prev;
	files_struct *fsp;
	int pnum;
	uint16 vuid;
	uint16 device_state;
	uint16 priority;
	fstring name;
	fstring pipe_srv_name;

	/* remote, server-side rpc redirection */
	struct msrpc_local *m;
} pipes_struct;

/* if a kernel does support oplocks then a structure of the following
   typee is used to describe how to interact with the kernel */
struct kernel_oplocks {
	BOOL (*receive_message)(fd_set *fds, char *buffer, int buffer_len);
	BOOL (*set_oplock)(files_struct *fsp, int oplock_type);
	void (*release_oplock)(files_struct *fsp);
	BOOL (*parse_message)(char *msg_start, int msg_len, SMB_INO_T *inode, SMB_DEV_T *dev);
	BOOL (*msg_waiting)(fd_set *fds);
	int notification_fd;
};


#define FSP_CAN_READ(fsp, want_exec) \
	((fsp)->can_read || (want_exec && (fsp)->can_execute))
#define CHECK_READ(fsp, in_packet) \
	if (!FSP_CAN_READ(fsp, SMB_IS_EXECUTE(in_packet))) \
		return(ERROR_DOS(ERRDOS,ERRbadaccess))


#include "smbd_proto.h"

extern GENERIC_MAPPING file_generic_mapping;
extern uint32 global_client_caps;
extern BOOL global_oplock_break;

/* 
 * variables used to override error codes. 
 */
extern int unix_ERR_class;
extern int unix_ERR_code;
extern NTSTATUS unix_ERR_ntstatus;


#endif
