/* 
   Unix SMB/Netbios implementation.
   Version 1.9.
   VFS structures and parameters
   Copyright (C) Tim Potter 1999
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#ifndef _VFS_H
#define _VFS_H

#include "smbd.h"

/* Avoid conflict with an AIX include file */

#ifdef vfs_ops
#undef vfs_ops
#endif

/* VFS operations structure */

struct vfs_ops {

    /* Disk operations */
    
    int (*connect)(struct connection_struct *conn, const char *service, const char *user);
    void (*disconnect)(struct connection_struct *conn);
    SMB_BIG_UINT (*disk_free)(char *path, BOOL small_query, SMB_BIG_UINT *bsize, 
			      SMB_BIG_UINT *dfree, SMB_BIG_UINT *dsize);
    
    /* Directory operations */

    DIR *(*opendir)(const char *fname);
    struct dirent *(*readdir)(DIR *dirp);
    int (*mkdir)(char *path, mode_t mode);
    int (*rmdir)(char *path);
    int (*closedir)(struct connection_struct *conn, DIR *dir);
    void (* rewinddir) (struct connection_struct *conn, DIR *dirp);
    
    /* File operations */
    
    int (*open)(const char *fname, int flags, mode_t mode);
    int (*close)(files_struct *fsp, int fd);
    ssize_t (*read)(int fd, char *data, size_t n);
    ssize_t (*write)(int fd, char *data, size_t n);
    SMB_OFF_T (*lseek)(int filedes, SMB_OFF_T offset, int whence);
    int (*rename)(char *old, char *new);
    int (*fsync)(int fd);
    int (*stat)(const char *fname, SMB_STRUCT_STAT *sbuf);
    int (*fstat)(int fd, SMB_STRUCT_STAT *sbuf);
    int (*lstat)(const char *path, SMB_STRUCT_STAT *sbuf);
    int (*unlink)(const char *path);
    int (*chmod)(const char *path, mode_t mode);
    int (*chown)(struct connection_struct *conn, const char *path, uid_t uid, gid_t gid);
    int (*utime)(char *path, struct utimbuf *times);
	int (*ftruncate)(int fd, SMB_OFF_T offset);
	BOOL (*lock)(int fd, int op, SMB_OFF_T offset, SMB_OFF_T count, int type);
};


#define vfs_stat(conn, name, st) \
	((conn)->vfs_ops->stat((name), (st)))
#define vfs_chmod(conn, name, perms) \
	((conn)->vfs_ops->chmod((name), (perms)))
#define vfs_chown(conn, name, uid, gid) \
	((conn)->vfs_ops->chown((conn), (name), (uid), (gid)))
#define vfs_fstat(fsp, fd, st) \
	((fsp)->conn->vfs_ops->fstat((fd), (st)))
#define vfs_mkdir(conn, name, mode) \
	((conn)->vfs_ops->mkdir((name), (mode)))
#define vfs_rmdir(conn, name) \
	((conn)->vfs_ops->rmdir(name))
#define SMB_VFS_LSTAT(conn, path, sbufp) \
	((conn)->vfs_ops->lstat((path), (sbufp)))
#define SMB_VFS_READLINK(conn, path, res, s) \
	sys_readlink((path), (res), (s))
#define SMB_VFS_UNLINK(conn, path) \
	((conn)->vfs_ops->unlink((path)))


#endif /* _VFS_H */
