/* 
   Unix SMB/Netbios implementation.
   Version 1.9.
   uid/user handling
   Copyright (C) Andrew Tridgell 1992-1998
   Copyright (C) Luke Kenneth Casson Leighton 1996-2001
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include "includes.h"
#include "debug.h"
#include "rpc_parse.h"
#include "others.h"


static struct uid_cache vcache;

void init_vuid(void)
{
	init_uid();
	vcache.entries = 0;
}

/****************************************************************************
  become the user of a connection number
****************************************************************************/
BOOL become_vuser(uint16 vuid)
{
	vuser_struct *vu;
	user_struct *vuser;
	BOOL ret;

	unbecome_vuser();
	vu = get_valid_vuser_struct(vuid);

	if (vu == NULL || vu->usr == NULL)
	{
		DEBUG(1, ("become_vuser: NULL valid user struct for 0x%x\n",
			  vuid));
		return False;
	}

	vuser = vu->usr;

#if 0
	if (!check_vuser_ok(&vcache, vuser, -1))
	{
		DEBUG(2,("become_vuser: check_vuser rejected access\n"));
		vuid_free_user_struct(vuser);
		return False;
	}
#endif

	ret = become_unix_sec_ctx(vuid, NULL, vuser->uid, vuser->gid,
	                           vuser->n_groups, vuser->groups);
	vuid_free_vuser_struct(vu);
	return ret;
}

/****************************************************************************
  unbecome a user 
****************************************************************************/
BOOL unbecome_vuser(void)
{
	return unbecome_to_initial_uid();
}

