/*
   Temporary Static Storage

   Copyright (C) Elrond <elrond at samba-tng dot org>  2004

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
/* MKPROTO: libsamba_proto.h		(do not remove this line) */


#include "includes.h"
#include <stdlib.h>
#include "libsamba.h"
#include "tng_misc.h"


#define ROTATE_SIZE (1 << 5)


typedef struct _tss_descriptor
{
	void **ptrs;
	size_t *sizes;
	size_t idx;
} TSSDescriptor;


static TSSDescriptor *global_tssd = NULL;

static TSSDescriptor *tss_get(void)
{
	TSSDescriptor *tss = global_tssd;

	if (tss)
	{
		DEBUG(7, ("tss-idx: %u\n", tss->idx));
		return tss;
	}

	tss = g_new(TSSDescriptor, 1);

	tss->sizes = g_new0(size_t, ROTATE_SIZE);
	tss->ptrs = g_new0(void *, ROTATE_SIZE);
	tss->idx = 0;

	global_tssd = tss;

	return tss;
}

void tss_addptr(void *ptr, size_t size)
{
	TSSDescriptor *tss = tss_get();
	size_t idx;

	if (! ptr)
		return;

	idx = (tss->idx + 1) % ROTATE_SIZE;
	tss->idx = idx;

	safe_free(tss->ptrs[idx]);
	tss->ptrs[idx] = ptr;
	tss->sizes[idx] = size;
}

void *tss_malloc(size_t size)
{
	TSSDescriptor *tss = tss_get();
	size_t idx;

	idx = (tss->idx + 1) % ROTATE_SIZE;
	tss->idx = idx;

	if (tss->sizes[idx] >= size)
		return tss->ptrs[idx];

	safe_free(tss->ptrs[idx]);
	tss->ptrs[idx] = g_malloc(size);
	tss->sizes[idx] = size;

	return tss->ptrs[idx];
}
