/* 
 *  Unix SMB/Netbios implementation.
 *  Version 1.9.
 *  RPC Pipe client / server routines
 *  Copyright (C) Andrew Tridgell              1992-1997,2000,
 *  Copyright (C) Luke Kenneth Casson Leighton 1996-1997,2000,
 *  Copyright (C) Paul Ashton                       1997,2000,
 *  Copyright (C) Elrond                            2000-2001
 *  
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/* MKPROTO: rpc_client_proto.h		(do not remove this line) */

#include "includes.h"
#include "libsamba.h"
#include "rpc_parse.h"
#include "rpc_client.h"
#include "nterr.h"


#define DECLARE_BASIC_VARS(MARSHALL_ST, UNMARSHALL_ST) \
	prs_struct data, rdata; \
	MARSHALL_ST q_o; \
	UNMARSHALL_ST r_o; \
	struct cli_connection *con = NULL
#define INIT_BASIC_VARS() \
	rpccli_prs_open(&data, &rdata)
#define DO_CON_PIPE_REQ(opcode) \
	rpc_con_pipe_req(con, (opcode), &data, &rdata)
#define report_error_code(opcodename, ntstatus) \
	DEBUG(0, (opcodename ": %s\n", \
		  get_nt_error_msg(ntstatus)))
#define clean_and_ret(val) \
	rpccli_close_prs(&data, &rdata); \
	\
	cli_connection_unlink(con); \
	\
	return val;



/****************************************************************************
do a server net tprt enum
****************************************************************************/
BOOL srv_net_srv_tprt_enum(const char *srv_name,
			   uint32 switch_value, SRV_TPRT_INFO_CTR * ctr,
			   uint32 preferred_len, ENUM_HND * hnd)
{
	DECLARE_BASIC_VARS(SRV_Q_NET_TPRT_ENUM, SRV_R_NET_TPRT_ENUM);
	BOOL p = True;
	NTSTATUS status = NT_STATUS_UNSUCCESSFUL;
	BOOL valid_enum = False;

	if (ctr == NULL || preferred_len == 0)
		return False;

	if (!cli_connection_init(srv_name, PIPE_SRVSVC, &con))
	{
		return False;
	}

	INIT_BASIC_VARS();

	DEBUG(4, ("SRV Net Server Transport Enum, level %d, enum:%8x\n",
		  switch_value, get_enum_hnd(hnd)));

	ctr->switch_value = switch_value;
	ctr->ptr_tprt_ctr = 1;
	ctr->tprt.info0.num_entries_read = 0;
	ctr->tprt.info0.ptr_tprt_info = 1;

	/* store the parameters */
	make_srv_q_net_tprt_enum(&q_o, srv_name,
				 switch_value, ctr, preferred_len, hnd);
	r_o.ctr = ctr;

	/* turn parameters into data stream */
	if (srv_io_q_net_tprt_enum("", &q_o, &data, 0)
	    && DO_CON_PIPE_REQ(SRV_NETTRANSPORTENUM)
	    && srv_io_r_net_tprt_enum("", &r_o, &rdata, 0))
	{
		status = WIN32_TO_NTSTATUS(r_o.status);
		if (! NT_STATUS_IS_OK(status))
		{
			report_error_code("SRV_NETTRANSPORTENUM", status);
			p = 0;
		}

		if (p && r_o.ctr->switch_value != switch_value)
		{
			/* different switch levels.  oops. */
			DEBUG(0,
			      ("SRV_NETTRANSPORTENUM: info class %d does not match request %d\n",
			       r_o.ctr->switch_value, switch_value));
			p = 0;
		}

		if (p)
		{
			/* ok, at last: we're happy. */
			valid_enum = True;
		}
	}

	clean_and_ret(valid_enum);
}

/****************************************************************************
do a server net conn enum
****************************************************************************/
BOOL srv_net_srv_conn_enum(char *srv_name, char *qual_name,
			   uint32 switch_value, SRV_CONN_INFO_CTR * ctr,
			   uint32 preferred_len, ENUM_HND * hnd)
{
	DECLARE_BASIC_VARS(SRV_Q_NET_CONN_ENUM, SRV_R_NET_CONN_ENUM);
	BOOL p = True;
	NTSTATUS status = NT_STATUS_UNSUCCESSFUL;
	BOOL valid_enum = False;

	if (ctr == NULL || preferred_len == 0)
		return False;

	if (!cli_connection_init(srv_name, PIPE_SRVSVC, &con))
	{
		return False;
	}

	INIT_BASIC_VARS();

	DEBUG(4, ("SRV Net Server Connection Enum %s), level %d, enum:%8x\n",
		  qual_name, switch_value, get_enum_hnd(hnd)));

	ctr->switch_value = switch_value;
	ctr->ptr_conn_ctr = 1;
	ctr->conn.info0.num_entries_read = 0;
	ctr->conn.info0.ptr_conn_info = 1;

	/* store the parameters */
	make_srv_q_net_conn_enum(&q_o, srv_name, qual_name,
				 switch_value, ctr, preferred_len, hnd);
	r_o.ctr = ctr;

	/* turn parameters into data stream */
	if (srv_io_q_net_conn_enum("", &q_o, &data, 0)
	    && DO_CON_PIPE_REQ(SRV_NETCONNENUM)
	    && srv_io_r_net_conn_enum("", &r_o, &rdata, 0))
	{
		status = WIN32_TO_NTSTATUS(r_o.status);
		if (! NT_STATUS_IS_OK(status))
		{
			report_error_code("SRV_NETCONNENUM", status);
			p = 0;
		}

		if (p && r_o.ctr->switch_value != switch_value)
		{
			/* different switch levels.  oops. */
			DEBUG(0,
			      ("SRV_NETCONNENUM: info class %d does not match request %d\n",
			       r_o.ctr->switch_value, switch_value));
			p = 0;
		}

		if (p)
		{
			/* ok, at last: we're happy. */
			valid_enum = True;
		}
	}

	clean_and_ret(valid_enum);
}

/****************************************************************************
do a server net sess enum
****************************************************************************/
BOOL srv_net_srv_sess_enum(char *srv_name, char *qual_name, char *user_name,
			   uint32 switch_value, SRV_SESS_INFO_CTR * ctr,
			   uint32 preferred_len, ENUM_HND * hnd)
{
	DECLARE_BASIC_VARS(SRV_Q_NET_SESS_ENUM, SRV_R_NET_SESS_ENUM);
	BOOL p = True;
	NTSTATUS status = NT_STATUS_UNSUCCESSFUL;
	BOOL valid_enum = False;

	if (ctr == NULL || preferred_len == 0)
		return False;

	if (!cli_connection_init(srv_name, PIPE_SRVSVC, &con))
	{
		return False;
	}

	INIT_BASIC_VARS();

	DEBUG(4, ("SRV Net Session Enum, level %d, enum:%8x\n",
		  switch_value, get_enum_hnd(hnd)));

	ctr->switch_value = switch_value;
	ctr->ptr_sess_ctr = 1;
	ctr->sess.info0.num_entries_read = 0;
	ctr->sess.info0.ptr_sess_info = 1;

	/* store the parameters */
	make_srv_q_net_sess_enum(&q_o, srv_name, qual_name, user_name,
				 switch_value, ctr, preferred_len, hnd);
	r_o.ctr = ctr;

	/* turn parameters into data stream */
	if (srv_io_q_net_sess_enum("", &q_o, &data, 0)
	    && DO_CON_PIPE_REQ(SRV_NETSESSENUM)
	    && srv_io_r_net_sess_enum("", &r_o, &rdata, 0))
	{
		status = WIN32_TO_NTSTATUS(r_o.status);
		if (! NT_STATUS_IS_OK(status))
		{
			report_error_code("SRV_NETSESSENUM", status);
			p = 0;
		}

		if (p && r_o.ctr->switch_value != switch_value)
		{
			/* different switch levels.  oops. */
			DEBUG(0,
			      ("SRV_NETSESSENUM: info class %d does not match request %d\n",
			       r_o.ctr->switch_value, switch_value));
			p = 0;
		}

		if (p)
		{
			/* ok, at last: we're happy. */
			valid_enum = True;
		}
	}

	clean_and_ret(valid_enum);
}

/****************************************************************************
do a server net share enum
****************************************************************************/
BOOL srv_net_srv_share_enum(char *srv_name,
			    uint32 switch_value, SRV_SHARE_INFO_CTR * ctr,
			    uint32 preferred_len, ENUM_HND * hnd)
{
	DECLARE_BASIC_VARS(SRV_Q_NET_SHARE_ENUM, SRV_R_NET_SHARE_ENUM);
	BOOL p = True;
	NTSTATUS status = NT_STATUS_UNSUCCESSFUL;
	BOOL valid_enum = False;

	if (ctr == NULL || preferred_len == 0)
		return False;

	if (!cli_connection_init(srv_name, PIPE_SRVSVC, &con))
	{
		return False;
	}

	INIT_BASIC_VARS();

	DEBUG(4, ("SRV Get Share Info, level %d, enum:%8x\n",
		  switch_value, get_enum_hnd(hnd)));

	q_o.share_level = switch_value;

	ctr->switch_value = switch_value;
	ctr->ptr_share_ctr = 1;
	ctr->share.info1.num_entries_read = 0;
	ctr->share.info1.ptr_share_info = 0;

	/* store the parameters */
	make_srv_q_net_share_enum(&q_o, srv_name,
				  switch_value, ctr, preferred_len, hnd);
	r_o.ctr = ctr;

	/* turn parameters into data stream */
	if (srv_io_q_net_share_enum("", &q_o, &data, 0)
	    && DO_CON_PIPE_REQ(SRV_NETSHAREENUM)
	    && srv_io_r_net_share_enum("", &r_o, &rdata, 0))
	{
		status = WIN32_TO_NTSTATUS(r_o.status);
		if (! NT_STATUS_IS_OK(status))
		{
			report_error_code("SRV_NETSHAREENUM", status);
			p = 0;
		}

		if (p && r_o.ctr->switch_value != switch_value)
		{
			/* different switch levels.  oops. */
			DEBUG(0,
			      ("SRV_NETSHAREENUM: info class %d does not match request %d\n",
			       r_o.ctr->switch_value, switch_value));
			p = 0;
		}

		if (p)
		{
			/* ok, at last: we're happy. */
			valid_enum = True;
		}
	}

	clean_and_ret(valid_enum);
}

/****************************************************************************
do a share get info
****************************************************************************/
uint32 srv_net_srv_share_get_info(const char *srv_name,
				  const char *share_name,
				  uint32 info_level, SHARE_INFO_CTR * ctr)
{
	DECLARE_BASIC_VARS(SRV_Q_NET_SHARE_GET_INFO, SRV_R_NET_SHARE_GET_INFO);
	UNISTR2 uni_srv_name;
	UNISTR2 uni_share_name;
	NTSTATUS status = NT_STATUS_UNSUCCESSFUL;

	if (srv_name == NULL || share_name == NULL || ctr == NULL)
		return NT_STATUS_INVALID_PARAMETER;

	if (!cli_connection_init(srv_name, PIPE_SRVSVC, &con))
		return status;

	INIT_BASIC_VARS();
	ZERO_STRUCT(r_o);

	DEBUG(4, ("SRV Get Share Info, share:%s, level %d\n",
		  share_name, info_level));

	make_buf_unistr2(&uni_srv_name, NULL, srv_name);
	make_buf_unistr2(&uni_share_name, NULL, share_name);

	/* store the parameters */
	make_srv_q_net_share_get_info(&q_o, &uni_srv_name, &uni_share_name,
				      info_level);
	ZERO_STRUCTP(ctr);
	r_o.ctr = ctr;

	/* turn parameters into data stream */
	if (srv_io_q_net_share_get_info("", &q_o, &data, 0)
	    && DO_CON_PIPE_REQ(SRV_NETSHAREGETINFO)
	    && srv_io_r_net_share_get_info("", &r_o, &rdata, 0))
	{
		status = WIN32_TO_NTSTATUS(r_o.status);
		if (! NT_STATUS_IS_OK(status))
			report_error_code("SRV_NETSHAREGETINFO", status);
	}

	clean_and_ret(status);
}

/****************************************************************************
do a share get del
****************************************************************************/
uint32 srv_net_srv_share_del(const char *srv_name,
			     const char *share_name)
{
	DECLARE_BASIC_VARS(SRV_Q_NET_SHARE_DEL, SRV_R_NET_SHARE_DEL);
	UNISTR2 uni_srv_name;
	UNISTR2 uni_share_name;
	NTSTATUS status = NT_STATUS_UNSUCCESSFUL;

	if (srv_name == NULL || share_name == NULL)
		return NT_STATUS_INVALID_PARAMETER;

	if (!cli_connection_init(srv_name, PIPE_SRVSVC, &con))
		return status;

	INIT_BASIC_VARS();
	ZERO_STRUCT(r_o);

	DEBUG(4, ("SRV Del Share, server:%s share:%s\n",
		  srv_name, share_name));

	make_unistr2(&uni_srv_name, srv_name, strlen(srv_name) + 1);
	make_unistr2(&uni_share_name, share_name, strlen(share_name) + 1);

	/* store the parameters */
	make_srv_q_net_share_del(&q_o, &uni_srv_name, &uni_share_name);

	/* turn parameters into data stream */
	if (srv_io_q_net_share_del("", &q_o, &data, 0)
	    && DO_CON_PIPE_REQ(SRV_NETSHAREDEL)
	    && srv_io_r_net_share_del("", &r_o, &rdata, 0))
	{
	 	status = WIN32_TO_NTSTATUS(r_o.status);
		if (! NT_STATUS_IS_OK(status))
			report_error_code("SRV_NETSHAREDEL", status);
	}

	clean_and_ret(status);
}

/****************************************************************************
do a share get del
****************************************************************************/
uint32 srv_net_share_add(const UNISTR2 *uni_srv_name,
			 uint32 info_level,
			 const SHARE_INFO_CTR * ctr, uint32 *parm_error)
{
	DECLARE_BASIC_VARS(SRV_Q_NET_SHARE_ADD, SRV_R_NET_SHARE_ADD);
	char *srv_name;
	NTSTATUS status = NT_STATUS_UNSUCCESSFUL;

	if (!uni_srv_name || !ctr || !parm_error)
		return NT_STATUS_INVALID_PARAMETER;

	srv_name = unistr2_to_ascii(NULL, uni_srv_name, 0);
	if (!cli_connection_init(srv_name, PIPE_SRVSVC, &con))
	{
		safe_free(srv_name);
		return status;
	}
	safe_free(srv_name);

	INIT_BASIC_VARS();
	ZERO_STRUCT(q_o);
	ZERO_STRUCT(r_o);

	q_o.ptr_srv_name = 1;
	copy_unistr2(&q_o.uni_srv_name, uni_srv_name);
	q_o.info_level = info_level;
	q_o.ctr = *ctr;
	q_o.parm_error = *parm_error;

	/* turn parameters into data stream */
	if (srv_io_q_net_share_add("", &q_o, &data, 0)
	    && DO_CON_PIPE_REQ(SRV_NET_SHARE_ADD)
	    && srv_io_r_net_share_add("", &r_o, &rdata, 0))
	{
	 	status = WIN32_TO_NTSTATUS(r_o.status);
		if (! NT_STATUS_IS_OK(status))
			report_error_code("SRV_NET_SHARE_ADD", status);
	 	*parm_error = r_o.parm_error;
	}

	clean_and_ret(status);
}

/****************************************************************************
do a server net file enum
****************************************************************************/
BOOL srv_net_srv_file_enum(char *srv_name, char *qual_name, uint32 file_id,
			   uint32 switch_value, SRV_FILE_INFO_CTR * ctr,
			   uint32 preferred_len, ENUM_HND * hnd)
{
	DECLARE_BASIC_VARS(SRV_Q_NET_FILE_ENUM, SRV_R_NET_FILE_ENUM);
	BOOL p = True;
	NTSTATUS status = NT_STATUS_UNSUCCESSFUL;
	BOOL valid_enum = False;

	if (ctr == NULL || preferred_len == 0)
		return False;

	if (!cli_connection_init(srv_name, PIPE_SRVSVC, &con))
	{
		return False;
	}

	INIT_BASIC_VARS();

	DEBUG(4, ("SRV Get File Info level %d, enum:%8x\n",
		  switch_value, get_enum_hnd(hnd)));

	q_o.file_level = switch_value;

	ctr->switch_value = switch_value;
	ctr->ptr_file_ctr = 1;
	ctr->file.info3.num_entries_read = 0;
	ctr->file.info3.ptr_file_info = 1;

	/* store the parameters */
	make_srv_q_net_file_enum(&q_o, srv_name, qual_name, file_id,
				 switch_value, ctr, preferred_len, hnd);
	r_o.ctr = ctr;

	/* turn parameters into data stream */
	if (srv_io_q_net_file_enum("", &q_o, &data, 0)
	    && DO_CON_PIPE_REQ(SRV_NETFILEENUM)
	    && srv_io_r_net_file_enum("", &r_o, &rdata, 0))
	{
		status = WIN32_TO_NTSTATUS(r_o.status);
		if (! NT_STATUS_IS_OK(status))
		{
			report_error_code("SRV_NETFILEENUM", status);
			p = 0;
		}

		if (p && r_o.ctr->switch_value != switch_value)
		{
			/* different switch levels.  oops. */
			DEBUG(0,
			      ("SRV_NETFILEENUM: info class %d does not match request %d\n",
			       r_o.ctr->switch_value, switch_value));
			p = 0;
		}

		if (p)
		{
			/* ok, at last: we're happy. */
			valid_enum = True;
		}
	}

	clean_and_ret(valid_enum);
}

/****************************************************************************
do a server get info 
****************************************************************************/
BOOL srv_net_srv_get_info(char *srv_name, uint32 switch_value,
			  SRV_INFO_CTR * ctr)
{
	DECLARE_BASIC_VARS(SRV_Q_NET_SRV_GET_INFO, SRV_R_NET_SRV_GET_INFO);
	BOOL p = True;
	NTSTATUS status = NT_STATUS_UNSUCCESSFUL;
	BOOL valid_info = False;

	if (switch_value == 0 || ctr == NULL)
		return False;

	if (!cli_connection_init(srv_name, PIPE_SRVSVC, &con))
	{
		return False;
	}

	INIT_BASIC_VARS();

	DEBUG(4, ("SRV Get Server Info level %d\n", switch_value));

	/* store the parameters */
	make_srv_q_net_srv_get_info(&q_o, srv_name, switch_value);
	r_o.ctr = ctr;

	/* turn parameters into data stream */
	if (srv_io_q_net_srv_get_info("", &q_o, &data, 0)
	    && DO_CON_PIPE_REQ(SRV_NET_SRV_GET_INFO)
	    && srv_io_r_net_srv_get_info("", &r_o, &rdata, 0))
	{
		status = WIN32_TO_NTSTATUS(r_o.status);
		if (! NT_STATUS_IS_OK(status))
		{
			report_error_code("SRV_NET_SRV_GET_INFO", status);
			p = 0;
		}

		if (p && r_o.ctr->switch_value != q_o.switch_value)
		{
			/* different switch levels.  oops. */
			DEBUG(0,
			      ("SRV_NET_SRV_GET_INFO: info class %d does not match request %d\n",
			       r_o.ctr->switch_value, q_o.switch_value));
			p = 0;
		}

		if (p)
		{
			/* ok, at last: we're happy. */
			valid_info = True;
		}
	}

	clean_and_ret(valid_info);
}

/****************************************************************************
get server time
****************************************************************************/
BOOL srv_net_remote_tod(char *srv_name, TIME_OF_DAY_INFO * tod)
{
	DECLARE_BASIC_VARS(SRV_Q_NET_REMOTE_TOD, SRV_R_NET_REMOTE_TOD);
	BOOL p = True;
	NTSTATUS status = NT_STATUS_UNSUCCESSFUL;
	BOOL valid_info = False;

	if (tod == NULL)
		return False;

	if (!cli_connection_init(srv_name, PIPE_SRVSVC, &con))
	{
		return False;
	}

	INIT_BASIC_VARS();

	DEBUG(4, ("SRV Remote TOD (%s)\n", srv_name));

	/* store the parameters */
	make_srv_q_net_remote_tod(&q_o, srv_name);
	r_o.tod = tod;

	/* turn parameters into data stream */
	if (srv_io_q_net_remote_tod("", &q_o, &data, 0)
	    && DO_CON_PIPE_REQ(SRV_NET_REMOTE_TOD)
	    && srv_io_r_net_remote_tod("", &r_o, &rdata, 0))
	{
		status = WIN32_TO_NTSTATUS(r_o.status);
		if (! NT_STATUS_IS_OK(status))
		{
			report_error_code("SRV_NET_REMOTE_TOD", status);
			p = False;
		}

		if (p)
		{
			valid_info = True;
		}
	}

	clean_and_ret(valid_info);
}
