/* 
 *  Unix SMB/Netbios implementation.
 *  Version 1.9.
 *  RPC Pipe client / server routines
 *  Copyright (C) Andrew Tridgell              1992-1997,
 *  Copyright (C) Luke Kenneth Casson Leighton 1996-1997,
 *  Copyright (C) Paul Ashton                       1997.
 *  
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/* MKPROTO: rpc_client_proto.h		(do not remove this line) */

#include "includes.h"
#include "debug.h"
#include "rpc_parse.h"
#include "rpc_client.h"
#include "nterr.h"


/****************************************************************************
do a WKS Open Policy
****************************************************************************/
BOOL wks_query_info( char *srv_name, uint32 switch_value,
			WKS_INFO_100 *wks100)
{
	prs_struct rbuf;
	prs_struct buf; 
	WKS_Q_QUERY_INFO q_o;
	WKS_R_QUERY_INFO r_o;
	BOOL p;
	BOOL valid_info = False;
	struct cli_connection *con = NULL;

	if (wks100 == NULL) return False;

	if (!cli_connection_init(srv_name, PIPE_WKSSVC, &con))
	{
		return False;
	}

	prs_init(&buf, MARSHALL);
	prs_init(&rbuf, UNMARSHALL);

	DEBUG(4,("WKS Query Info\n"));

	/* store the parameters */
	make_wks_q_query_info(&q_o, srv_name, switch_value);

	/* turn parameters into data stream */
	if (wks_io_q_query_info("", &q_o, &buf, 0) &&
	    rpc_con_pipe_req(con, WKS_QUERY_INFO, &buf, &rbuf))
	{
		r_o.wks100 = wks100;

		wks_io_r_query_info("", &r_o, &rbuf, 0);
		p = rbuf.offset != 0;

		if (p && r_o.status != 0)
		{
			/* report error code */
			DEBUG(0,("WKS_R_QUERY_INFO: %s\n", get_nt_error_msg(r_o.status)));
			p = False;
		}

		if (p)
		{
			valid_info = True;
		}
	}

	rpccli_close_prs(&buf, &rbuf);

	cli_connection_unlink(con);

	return valid_info;
}

uint32 wks_user_enum(const char *srv_name, uint32 switch_value,
		     WKS_LOGGED_IN_USER **users, uint32 *count)
{
	prs_struct rbuf;
	prs_struct buf; 
	WKS_Q_USER_ENUM q_o;
	WKS_R_USER_ENUM r_o;
	uint32 status = NT_STATUS_UNSUCCESSFUL;
	struct cli_connection *con = NULL;

	if (srv_name==NULL)
		return NT_STATUS_INVALID_PARAMETER;

	if (!cli_connection_init(srv_name, PIPE_WKSSVC, &con))
	{
		return status;
	}

	prs_init(&buf, MARSHALL);
	prs_init(&rbuf, UNMARSHALL);
	ZERO_STRUCT(r_o);

	DEBUG(4,("WKS Query Info\n"));

	/* store the parameters */
	make_wks_q_user_enum(&q_o, srv_name, switch_value);

	/* turn parameters into data stream */
	if (wks_io_q_user_enum("", &q_o, &buf, 0)
	    && rpc_con_pipe_req(con, WKS_USER_ENUM, &buf, &rbuf)
	    && wks_io_r_user_enum("", &r_o, &rbuf, 0))
	{
		status = r_o.status;
		(*count) = r_o.count1;
		(*users) = r_o.users;
	}

	rpccli_close_prs(&buf, &rbuf);

	cli_connection_unlink(con);

	return status;
}


NTSTATUS wks_unjoin(const char *srv_name)
{
	prs_struct rbuf;
	prs_struct buf; 
	WKS_Q_UNJOIN q_o;
	WKS_R_UNJOIN r_o;
	uint32 status = NT_STATUS_UNSUCCESSFUL;
	struct cli_connection *con = NULL;

	if (srv_name==NULL)
		return NT_STATUS_INVALID_PARAMETER;

	if (!cli_connection_init(srv_name, PIPE_WKSSVC, &con))
	{
		return status;
	}

	prs_init(&buf, MARSHALL);
	prs_init(&rbuf, UNMARSHALL);

	ZERO_STRUCT(q_o);
	make_buf_unistr2(&(q_o.uni_srv_name), &(q_o.ptr_srv_name), srv_name);
	q_o.unk2 = 4;

	if (wks_io_q_unjoin("", &q_o, &buf, 0)
	    && rpc_con_pipe_req(con, WKS_UNJOIN, &buf, &rbuf)
	    && wks_io_r_unjoin("", &r_o, &rbuf, 0))
	{
		status = WIN32_TO_NTSTATUS(r_o.status);
	}

	rpccli_close_prs(&buf, &rbuf);

	cli_connection_unlink(con);

	return status;
}
