/* 
 *  Unix SMB/Netbios implementation.
 *  Version 1.9.
 *  RPC Pipe client / server routines
 *  Copyright (C) Andrew Tridgell              1992-2000,
 *  Copyright (C) Luke Kenneth Casson Leighton 1996-2000,
 *  Copyright (C) Jean Franois Micouleau      1998-2000,
 *  Copyright (C) Elrond                            2000.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/* MKPROTO: rpc_parse_proto.h		(do not remove this line) */
 
#include "includes.h"
#include "tng_misc.h"
#include "rpc_parse.h"


/*******************************************************************
********************************************************************/  
BOOL make_eventlog_q_open(EVENTLOG_Q_OPEN *q_u, const char *journal, char *unk)
{
	int len_journal = journal != NULL ? strlen(journal) : 0;
	int len_unk = unk != NULL ? strlen(unk) : 0;

	q_u->ptr0=0x1;
	q_u->unk0=0x5c;
	q_u->unk1=0x01;

	make_uni_hdr(&(q_u->hdr_source), len_journal);
	make_unistr2(&(q_u->uni_source), journal, len_journal);
	
	make_uni_hdr(&(q_u->hdr_unk), len_unk);
	make_unistr2(&(q_u->uni_unk), unk, len_unk);
	
	q_u->unk6=0x01; /* one of these is an access mask! */
	q_u->unk7=0x01; /* one of these is an access mask! */

	return True;
}

RPC_IO_DECLARE(eventlog_io_q_open, EVENTLOG_Q_OPEN, q_u)
{
	RPC_MARSHALLER_INTRO(q_u);

	prs_uint32("ptr0", ps, depth, &(q_u->ptr0));

	prs_uint16("unk0", ps, depth, &(q_u->unk0));
	prs_uint16("unk1", ps, depth, &(q_u->unk1));

	RPC_MARSH_SUBCALL(smb_io_unihdr, q_u, hdr_source);
	smb_io_unistr2("uni_source", &(q_u->uni_source),
		       q_u->hdr_source.buffer, ps, depth);
	prs_align(ps);

	RPC_MARSH_SUBCALL(smb_io_unihdr, q_u, hdr_unk);
	smb_io_unistr2("uni_unk", &(q_u->uni_unk),
		       q_u->hdr_unk.buffer, ps, depth);
	prs_align(ps);

	prs_uint32("unk6", ps, depth, &(q_u->unk6));
	prs_uint32("unk7", ps, depth, &(q_u->unk7));

	return True;
}

RPC_IO_DECLARE(eventlog_io_r_open, EVENTLOG_R_OPEN, r_u)
{
	RPC_MARSHALLER_INTRO(r_u);
	RPC_MARSH_SUBCALL(smb_io_pol_hnd, r_u, pol);
	prs_uint32("status", ps, depth, &(r_u->status));

	return True;
}

/*******************************************************************
********************************************************************/  
BOOL make_eventlog_q_close(EVENTLOG_Q_CLOSE *q_u, POLICY_HND *pol)
{
	if ((q_u == NULL) || (pol == NULL))
	{
		return False;
	}

	q_u->pol = *pol;

	return True;
}

RPC_IO_DECLARE(eventlog_io_q_close, EVENTLOG_Q_CLOSE, q_u)
{
	RPC_MARSHALLER_INTRO(q_u);
	RPC_MARSH_SUBCALL(smb_io_pol_hnd, q_u, pol);

	return True;
}

RPC_IO_DECLARE(eventlog_io_r_close, EVENTLOG_R_CLOSE, r_u)
{
	RPC_MARSHALLER_INTRO(r_u);
	RPC_MARSH_SUBCALL(smb_io_pol_hnd, r_u, pol);
	prs_uint32("status", ps, depth, &(r_u->status));

	return True;
}

/*******************************************************************
********************************************************************/  
BOOL make_eventlog_q_numofeventlogrec(EVENTLOG_Q_NUMOFEVENTLOGREC *q_u, POLICY_HND *pol)
{
	if ((q_u == NULL) || (pol == NULL))
	{
		return False;
	}

	q_u->pol = *pol;

	return True;
}

RPC_IO_DECLARE(eventlog_io_q_numofeventlogrec, EVENTLOG_Q_NUMOFEVENTLOGREC, q_u)
{
	RPC_MARSHALLER_INTRO(q_u);
	RPC_MARSH_SUBCALL(smb_io_pol_hnd, q_u, pol);

	return True;
}

RPC_IO_DECLARE(eventlog_io_r_numofeventlogrec, EVENTLOG_R_NUMOFEVENTLOGREC, r_u)
{
	RPC_MARSHALLER_INTRO(r_u);
	prs_uint32("number", ps, depth, &(r_u->number));
	prs_uint32("status", ps, depth, &(r_u->status));

	return True;
}

/*******************************************************************
********************************************************************/  
BOOL make_eventlog_q_readeventlog(EVENTLOG_Q_READEVENTLOG *q_u, POLICY_HND *pol,
                                  uint32 flags, uint32 offset, uint32 number_of_bytes)
{
	if ((q_u == NULL) || (pol == NULL))
	{
		return False;
	}

	q_u->pol = *pol;
	q_u->flags = flags;
	q_u->offset = offset;
	q_u->number_of_bytes = number_of_bytes;

	return True;
}

RPC_IO_DECLARE(eventlog_io_q_readeventlog, EVENTLOG_Q_READEVENTLOG, q_u)
{
	RPC_MARSHALLER_INTRO(q_u);
	RPC_MARSH_SUBCALL(smb_io_pol_hnd, q_u, pol);
	prs_uint32("flags",           ps, depth, &(q_u->flags));
	prs_uint32("offset",          ps, depth, &(q_u->offset));
	prs_uint32("number_of_bytes", ps, depth, &(q_u->number_of_bytes));

	return True;
}

static RPC_IO_DECLARE(eventlog_io_eventlog, EVENTLOGRECORD, ev)
{
	uint8 old_align;
	RPC_MARSHALLER_INTRO(ev);
	prs_uint32("size", ps, depth, &(ev->size));	
	prs_uint32("reserved", ps, depth, &(ev->reserved));
	prs_uint32("recordnumber", ps, depth, &(ev->recordnumber));
	prs_uint32("creationtime", ps, depth, &(ev->creationtime));
	prs_uint32("writetime", ps, depth, &(ev->writetime));
	prs_uint32("eventnumber", ps, depth, &(ev->eventnumber));
	
	prs_uint16("eventtype", ps, depth, &(ev->eventtype));
	prs_uint16("num_of_strings", ps, depth, &(ev->num_of_strings));
	prs_uint16("category", ps, depth, &(ev->category));
	prs_uint16("reserved_flag", ps, depth, &(ev->reserved_flag));

	prs_uint32("closingrecord", ps, depth, &(ev->closingrecord));
	prs_uint32("stringoffset", ps, depth, &(ev->stringoffset));
	prs_uint32("sid_length", ps, depth, &(ev->sid_length));
	prs_uint32("sid_offset", ps, depth, &(ev->sid_offset));
	prs_uint32("data_length", ps, depth, &(ev->data_length));
	prs_uint32("data_offset", ps, depth, &(ev->data_offset));
	
	RPC_MARSH_SUBCALL(smb_io_unistr, ev, sourcename);	
	RPC_MARSH_SUBCALL(smb_io_unistr, ev, computername);

	old_align = prs_set_align(ps, 0);

	if (ev->sid_length!=0)
		RPC_MARSH_SUBCALL(smb_io_dom_sid, ev, sid);
	
	if (ev->num_of_strings!=0)
	{
		uint32 i;
		ev->strings = g_new(UNISTR *, ev->num_of_strings);
		if(! ev->strings)
			return False;
		for (i = 0; i < ev->num_of_strings; i++)
		{
			fstring temp;
			ev->strings[i] = g_new(UNISTR, 1);
			if (!ev->strings[i])
				return False;
			snprintf(temp, sizeof(temp), "string[%d]", i);
			smb_io_unistr(temp, ev->strings[i], ps, depth);
		}
	}
	else
		ev->strings = NULL;	

	if (ev->data_length)
		RPC_MARSH_SUBCALL(smb_io_unistr, ev, data);

	prs_uint32("size2", ps, depth, &(ev->size2));	

	prs_set_align(ps, old_align);

	return True;
}

RPC_IO_DECLARE(eventlog_io_r_readeventlog, EVENTLOG_R_READEVENTLOG, r_u)
{
	RPC_MARSHALLER_INTRO(r_u);
	prs_uint32("number_of_bytes", ps, depth, &(r_u->number_of_bytes));

	if (r_u->number_of_bytes!= 0)
		RPC_MARSH_SUBCALLP(eventlog_io_eventlog, r_u, event);

	prs_uint32("sent_size", ps, depth, &(r_u->sent_size));		
	prs_uint32("real_size", ps, depth, &(r_u->real_size));
	prs_uint32("status", ps, depth, &(r_u->status));	

	return True;
}
