/* 
   Unix SMB/Netbios implementation.
   Version 1.9.
   NT Domain Authentication SMB / MSRPC client
   Copyright (C) Andrew Tridgell 1994-1997
   Copyright (C) Luke Kenneth Casson Leighton 1996-1997
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include "includes.h"
#include "libsamba.h"
#include "rpc_client.h"
#include "rpcclient.h"
#include "nterr.h"

extern FILE* out_hnd;

/****************************************************************************
workstation get info query
****************************************************************************/
NTSTATUS cmd_wks_query_info(struct client_info *info, int argc, char *argv[])
{
	fstring dest_wks;
	WKS_INFO_100 ctr;
	uint32 info_level = 100;

	ZERO_STRUCT(ctr);

	fstrcpy(dest_wks, "\\\\");
	fstrcat(dest_wks, info->dest_host);
	strupper(dest_wks);

	if (argc > 1)
	{
		info_level = (uint32)strtol(argv[1], (char**)NULL, 10);
	}

	DEBUG(4,("cmd_wks_query_info: server:%s info level: %d\n",
				dest_wks, info_level));

	/* send info level: receive requested info.  hopefully. */
	if (!wks_query_info( dest_wks, info_level, &ctr))
	{
		DEBUG(5,("cmd_wks_query_info: query failed\n"));
		return NT_STATUS_UNSUCCESSFUL;
	}

	DEBUG(5,("cmd_wks_query_info: query succeeded\n"));

	display_wks_info_100(out_hnd, ACTION_HEADER   , &ctr);
	display_wks_info_100(out_hnd, ACTION_ENUMERATE, &ctr);
	display_wks_info_100(out_hnd, ACTION_FOOTER   , &ctr);

	return NT_STATUS_OK;
}

/****************************************************************************
****************************************************************************/
NTSTATUS cmd_wks_user_enum(struct client_info *info, int argc, char *argv[])
{
	fstring dest_wks;
	WKS_LOGGED_IN_USER *users;
	uint32 count, i;
	uint32 info_level = 1;
	NTSTATUS status;

	fstrcpy(dest_wks, "\\\\");
	fstrcat(dest_wks, info->dest_host);
	strupper(dest_wks);

	if (argc > 1)
	{
		info_level = (uint32)strtol(argv[1], (char**)NULL, 10);
	}

	DEBUG(4, ("cmd_wks_user_enum: server:%s info level: %d\n",
		  dest_wks, info_level));

	count = 0;

	status = wks_user_enum(dest_wks, info_level, &users, &count);

	for (i = 0; i < count; i++)
	{
		char *domain, *user, *logonserver, *unk;

		domain = unistr2_to_ascii(NULL, &users[i].uni_domain, 0);
		user = unistr2_to_ascii(NULL, &users[i].uni_username, 0);
		logonserver = unistr2_to_ascii(NULL, &users[i].uni_logonserver, 0);
		unk = unistr2_to_ascii(NULL, &users[i].uni_unk, 0);

		report(out_hnd, "%s\\%s (%s) `%s'\n", domain, user,
		      logonserver, unk);
		safe_free(domain);
		safe_free(user);
		safe_free(logonserver);
		safe_free(unk);
	}

	if (NT_STATUS_IS_ERR(status))
	{
		report(out_hnd, "failed\n");
	}

	return status;
}


NTSTATUS cmd_wks_unjoin(struct client_info *info, int argc, char *argv[])
{
	fstring dest_wks;
	NTSTATUS status;

	fstrcpy(dest_wks, "\\\\");
	fstrcat(dest_wks, info->dest_host);
	strupper(dest_wks);

	status = wks_unjoin(dest_wks);

	return status;
}
