/* 
 *  
 *
 *  Copyright (C) Sander Striker                    2000.
 *  Copyright (C) Elrond                            2000.
 *  
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "uuid.h"
#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>

signed32
uuid_compare(uuid_t *uuid1, uuid_t *uuid2, unsigned32 *status)
{
  assert (uuid1 != NULL);
  assert (uuid2 != NULL);
  assert (status != NULL);

  *status = uuid_s_ok;
  if (uuid1->time_low < uuid2->time_low)
  {
    return -1;
  }
  else if (uuid1->time_low > uuid2->time_low)
  {
    return 1;
  }

  if (uuid1->time_mid < uuid2->time_mid)
  {
    return -1;
  }
  else if (uuid1->time_mid > uuid2->time_mid)
  {
    return 1;
  }

  if (uuid1->time_hi_and_version < uuid2->time_hi_and_version)
  {
    return -1;
  }
  else if (uuid1->time_hi_and_version > uuid2->time_hi_and_version)
  {
    return 1;
  }

  if (uuid1->clock_seq_hi_and_reserved < uuid2->clock_seq_hi_and_reserved)
  {
    return -1;
  }
  else if (uuid1->clock_seq_hi_and_reserved > uuid2->clock_seq_hi_and_reserved)
  {
    return 1;
  }

  if (uuid1->clock_seq_low < uuid2->clock_seq_low)
  {
    return -1;
  }
  else if (uuid1->clock_seq_low > uuid2->clock_seq_low)
  {
    return 1;
  }

  if (uuid1->node[0] < uuid2->node[0])
  {
    return -1;
  }
  else if (uuid1->node[0] > uuid2->node[0])
  {
    return 1;
  }

  if (uuid1->node[1] < uuid2->node[1])
  {
    return -1;
  }
  else if (uuid1->node[1] > uuid2->node[1])
  {
    return 1;
  }

  if (uuid1->node[2] < uuid2->node[2])
  {
    return -1;
  }
  else if (uuid1->node[2] > uuid2->node[2])
  {
    return 1;
  }

  if (uuid1->node[3] < uuid2->node[3])
  {
    return -1;
  }
  else if (uuid1->node[3] > uuid2->node[3])
  {
    return 1;
  }

  if (uuid1->node[4] < uuid2->node[4])
  {
    return -1;
  }
  else if (uuid1->node[4] > uuid2->node[4])
  {
    return 1;
  }

  if (uuid1->node[5] < uuid2->node[5])
  {
    return -1;
  }
  else if (uuid1->node[5] > uuid2->node[5])
  {
    return 1;
  }

  /*
   * Above code is an unroll of the following loop:  
   * for (int n = 0; n < 6; n++)
   * {
   *   if (uuid1->node[n] < uuid2->node[n])
   *   {
   *     return -1;
   *   }
   *   else if (uuid1->node[n] < uuid2->node[n])
   *   {
   *     return -1;
   *   }
   * }
   */

  return 0;
}

void
uuid_create(uuid_t *uuid, unsigned32 *status)
{

}

void
uuid_create_nil(uuid_t *uuid, unsigned32 *status)
{
  assert (uuid != NULL);
  assert (status != NULL);

  memset(uuid, 0, sizeof(uuid_t));
  *status = uuid_s_ok;
}

boolean32
uuid_equal(uuid_t *uuid1, uuid_t *uuid2, unsigned32 *status)
{
  assert (uuid1 != NULL);
  assert (uuid2 != NULL);
  assert (status != NULL);

  *status = uuid_s_ok;
  return ((unsigned32 *)uuid1)[0] == ((unsigned32 *)uuid2)[0] &&
         ((unsigned32 *)uuid1)[1] == ((unsigned32 *)uuid2)[1] &&
         ((unsigned32 *)uuid1)[2] == ((unsigned32 *)uuid2)[2] &&
         ((unsigned32 *)uuid1)[3] == ((unsigned32 *)uuid2)[3];
  /*
   * Above return statement is the same as:
   * return uuid1->time_low == uuid2->time_low &&
   *        uuid1->time_mid == uuid2->time_mid &&
   *        uuid1->time_hi_and_version == uuid2->time_hi_and_version &&
   *        uuid1->clock_seq_hi_and_reserved == uuid2->clock_seq_hi_and_reserved &&
   *        uuid1->clock_seq_low == uuid2->clock_seq_low &&
   *        uuid1->node[0] == uuid2->node[0] &&
   *        uuid1->node[1] == uuid2->node[1] &&
   *        uuid1->node[2] == uuid2->node[2] &&
   *        uuid1->node[3] == uuid2->node[3] &&
   *        uuid1->node[4] == uuid2->node[4] &&
   *        uuid1->node[5] == uuid2->node[5];
   */
}

/*
 * Main structure and algorithm for the following function
 * is based on the uuid-code in e2fsprogs (1.18).
 */

void
uuid_from_string(unsigned_char_t *string_uuid, uuid_t *uuid, unsigned32 *status)
{
  int i;
  const char *cp;
  char buf[3];

  assert (string_uuid != NULL);
  assert (uuid != NULL);
  assert (status != NULL);

  if (strlen(string_uuid) != 36) {
    *status = -1;
    return;
  }
  for (i=0, cp = string_uuid; i <= 36; i++,cp++) {
    if ((i == 8) || (i == 13) || (i == 18) ||
	(i == 23))
      if (*cp == '-')
	continue;
    if (i== 36)
      if (*cp == 0)
	continue;
    if (!isxdigit(*cp)) {
      *status = -1;
      return;
    }
  }
  uuid->time_low = strtoul(string_uuid, NULL, 16);
  uuid->time_mid = strtoul(string_uuid+9, NULL, 16);
  uuid->time_hi_and_version = strtoul(string_uuid+14, NULL, 16);

  buf[2] = 0;

  cp = string_uuid+19;
  buf[0] = *cp++;
  buf[1] = *cp++;
  uuid->clock_seq_hi_and_reserved = strtoul(buf, NULL, 16);
  buf[0] = *cp++;
  buf[1] = *cp++;
  uuid->clock_seq_low = strtoul(buf, NULL, 16);

  cp = string_uuid+24;
  for (i=0; i < 6; i++) {
    buf[0] = *cp++;
    buf[1] = *cp++;
    uuid->node[i] = strtoul(buf, NULL, 16);
  }

  *status = uuid_s_ok;
}

#if 0
signed16
uuid_hash(uuid_t *uuid, unsigned32 *status)
{
  assert (uuid != NULL);
  assert (status != NULL);

}
#endif

boolean32
uuid_is_nil(uuid_t *uuid, unsigned32 *status)
{
  assert (uuid != NULL);
  assert (status != NULL);

  *status = uuid_s_ok;
  return ((unsigned32 *)uuid)[0] == 0 &&
         ((unsigned32 *)uuid)[1] == 0 &&
         ((unsigned32 *)uuid)[2] == 0 &&
         ((unsigned32 *)uuid)[3] == 0;
  /*
   * Above return statement is the same as:
   * return uuid->time_low == 0 &&
   *        uuid->time_mid == 0 &&
   *        uuid->time_hi_and_version == 0 &&
   *        uuid->clock_seq_hi_and_reserved == 0 &&
   *        uuid->clock_seq_low == 0 &&
   *        uuid->node[0] == 0 &&
   *        uuid->node[1] == 0 &&
   *        uuid->node[2] == 0 &&
   *        uuid->node[3] == 0 &&
   *        uuid->node[4] == 0 &&
   *        uuid->node[5] == 0;
   */
}

void
uuid_to_string(uuid_t *uuid, unsigned_char_t **string_uuid, unsigned32 *status)
{
  assert (uuid != NULL);
  assert (string_uuid != NULL);
  assert (status != NULL);

  *string_uuid = malloc(37);

  snprintf(*string_uuid, 37, "%08lx-%04x-%04x-%02x%02x-%02x%02x%02x%02x%02x%02x", 
          (unsigned long)uuid->time_low,
          uuid->time_mid,
          uuid->time_hi_and_version,
          uuid->clock_seq_hi_and_reserved,
          uuid->clock_seq_low,
          uuid->node[0],
          uuid->node[1],
          uuid->node[2],
          uuid->node[3],
          uuid->node[4],
          uuid->node[5]);
  *status = uuid_s_ok;
}

/* extensions to the original API */

void
uuid_copy(uuid_t *dest, const uuid_t *src, unsigned32 *status)
{
  assert (dest != NULL);
  assert (src != NULL);
  assert (status != NULL);

  ((unsigned32 *)dest)[0] = ((const unsigned32 *)src)[0];
  ((unsigned32 *)dest)[1] = ((const unsigned32 *)src)[1];
  ((unsigned32 *)dest)[2] = ((const unsigned32 *)src)[2];
  ((unsigned32 *)dest)[3] = ((const unsigned32 *)src)[3];

  /*
   * Above statements are the same as:
   * dest->time_low = src->time_low;
   * dest->time_mid = src->time_mid;
   * dest->time_hi_and_version = src->time_hi_and_version;
   * dest->clock_seq_hi_and_reserved = src->clock_seq_hi_and_reserved;
   * dest->clock_seq_low = src->clock_seq_low;
   * dest->node[0] = src->node[0];
   * dest->node[1] = src->node[1];
   * dest->node[2] = src->node[2];
   * dest->node[3] = src->node[3];
   * dest->node[4] = src->node[4];
   * dest->node[5] = src->node[5];
   */

  *status = uuid_s_ok;
}

